//// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
//// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
//// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
//// PARTICULAR PURPOSE.
////
//// Copyright (c) Microsoft Corporation. All rights reserved

#include "pch.h"
#include "ShapeRenderer.h"

using namespace concurrency;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::ApplicationModel;
using namespace Windows::ApplicationModel::Activation;
using namespace Windows::ApplicationModel::Core;
using namespace Windows::ApplicationModel::DataTransfer;
using namespace Windows::Foundation;
using namespace Windows::Graphics::Display;
using namespace Windows::Storage;
using namespace Windows::Storage::Pickers;
using namespace Windows::Storage::Streams;
using namespace Windows::UI::ApplicationSettings;
using namespace Windows::UI::Core;
using namespace Windows::UI::Input;
using namespace Windows::UI::Input::Inking;
using namespace Windows::UI::ViewManagement;
using namespace Windows::UI::Xaml::Controls;
using namespace Windows::UI::Xaml::Input;

ShapeRenderer::ShapeRenderer():
    m_backgroundColor(Windows::UI::Colors::White),
    m_backgroundBrush(nullptr),
    m_inkBrush(nullptr),
    m_currentBuffer(nullptr),
    m_previousBuffer(nullptr),
    m_inkStyle(nullptr),
    m_pointerId(-1),
    m_renderingMode(Bezier),
    m_manipulationMode(Windows::UI::Input::Inking::InkManipulationMode::Inking)
{
    m_numBuffers = 4;

    m_manipulationPoints = ref new Platform::Collections::Vector<Windows::Foundation::Point>();

    m_strokeBuilder = ref new Windows::UI::Input::Inking::InkStrokeBuilder();
    m_strokeContainer = ref new Windows::UI::Input::Inking::InkStrokeContainer();

    m_drawingAttributes = ref new Windows::UI::Input::Inking::InkDrawingAttributes();
    m_drawingAttributes->Color = Windows::UI::Colors::Black;
    m_drawingAttributes->PenTip = Windows::UI::Input::Inking::PenTipShape::Circle;
    m_drawingAttributes->Size = Windows::Foundation::Size(3.0f, 3.0f);
    m_drawingAttributes->IgnorePressure = true;
    m_drawingAttributes->FitToCurve = true;

    m_strokeBuilder->SetDefaultDrawingAttributes(m_drawingAttributes.Get());
}

bool ShapeRenderer::EventIsErase(_In_ Windows::UI::Xaml::Input::PointerRoutedEventArgs^ e)
{
    auto pointerDevice = e->GetCurrentPoint(nullptr)->PointerDevice;
    auto pointerProperties = e->GetCurrentPoint(nullptr)->Properties;

    bool rval = false
        || (pointerProperties->IsEraser)
        || (pointerDevice->PointerDeviceType == Windows::Devices::Input::PointerDeviceType::Pen && !pointerProperties->IsBarrelButtonPressed && pointerProperties->IsRightButtonPressed)
        || (pointerDevice->PointerDeviceType == Windows::Devices::Input::PointerDeviceType::Mouse && pointerProperties->IsRightButtonPressed);

    return rval;
}

bool ShapeRenderer::EventIsInk(_In_ Windows::UI::Xaml::Input::PointerRoutedEventArgs^ e)
{
    return true;
}

PointerEventType ShapeRenderer::GetPointerEventType(_In_ Windows::UI::Xaml::Input::PointerRoutedEventArgs^ e)
{
    PointerEventType type = PointerEventType::None;

    if (EventIsErase(e))
    {
        type = PointerEventType::Erase;
    }
    else if (EventIsInk(e))
    {
        type = PointerEventType::Ink;
    }   

    return type;
}

void ShapeRenderer::CreateBrush(
    _In_ ID2D1DeviceContext* d2dContext, 
    _In_ Windows::UI::Color color, 
    _Outptr_ ID2D1SolidColorBrush** brush)
{
    D2D1::ColorF d2dColor(color.R/255.0f, color.G/255.0f, color.B/255.0f);
    *brush = nullptr;
    DX::ThrowIfFailed(
        d2dContext->CreateSolidColorBrush(
        d2dColor,
        brush
        )
        );
}

void ShapeRenderer::CreateDeviceIndependentResources()
{
    DirectXBase::CreateDeviceIndependentResources();

    DX::ThrowIfFailed(
        m_d2dFactory->CreateStrokeStyle(
        D2D1::StrokeStyleProperties(
        D2D1_CAP_STYLE_ROUND,
        D2D1_CAP_STYLE_ROUND,
        D2D1_CAP_STYLE_ROUND,
        D2D1_LINE_JOIN_ROUND,
        1.0f,
        D2D1_DASH_STYLE_SOLID,
        0.0f),
        nullptr,
        0,
        &m_inkStyle));
}

void ShapeRenderer::CreateDeviceResources()
{
    DirectXBase::CreateDeviceResources();

    CreateBrush(m_d2dContext.Get(), m_backgroundColor, &m_backgroundBrush);
    CreateBrush(m_d2dContext.Get(), m_drawingAttributes->Color, &m_inkBrush);
}

void ShapeRenderer::CreateWindowSizeDependentResources()
{
    m_previousBuffer = nullptr;
    m_currentBuffer = nullptr;

    DirectXBase::CreateWindowSizeDependentResources();

    // Retrieve buffers 0 and m_numBuffers-1: we will need them in live rendering mode.
    //
    // In live rendering mode we need to copy the content of the last buffer we
    // presented into the new buffer we are going to present. This is more
    // efficient than redoing the whole bezier render and the previous parts
    // of the live render.
    // The next buffer that will be presented is always buffer 0, and the last
    // buffer presented is always buffer m_numBuffers - 1.
    DX::ThrowIfFailed(
        m_swapChain->GetBuffer(
        0,
        __uuidof(ID3D11Texture2D),
        &m_currentBuffer
        )
        );
    DX::ThrowIfFailed(
        m_swapChain->GetBuffer(
        m_numBuffers - 1,
        __uuidof(ID3D11Texture2D),
        &m_previousBuffer
        )
        );
}

void ShapeRenderer::Present()
{
    auto hr = S_OK;

    DXGI_PRESENT_PARAMETERS parameters = {0};
    parameters.DirtyRectsCount = 0;
    parameters.pDirtyRects = nullptr;
    parameters.pScrollRect = nullptr;
    parameters.pScrollOffset = nullptr;

    // In the interest of minimizing lagging between the rendered stroke and the
    // position of the inking device, we shall use immediate-mode presentation
    // (Present(0, ...)) while in live rendering. This way we can always present
    // the most up to date inking sample.
    // Since immediate mode presentation drains the battery we use non-immediate
    // mode presentation (Present(1, ...)) when we are not in live rendering mode.
    switch (m_renderingMode)
    {
    case Bezier:
        hr = m_swapChain->Present(1, 0);
        break;
    case Live:
        hr = m_swapChain->Present(0, 0);
        break;
    }

    // If the device was removed either by a disconnect or a driver upgrade, we 
    // must recreate all device resources.
    if (hr == DXGI_ERROR_DEVICE_REMOVED)
    {
        HandleDeviceLost();
    }
    else
    {
        DX::ThrowIfFailed(hr);
    }
}

void ShapeRenderer::UpdateForWindowSizeChange()
{
    // Release any extra references to Direct3D resources before attempting
    // to resize the swap chain.    

    DirectXBase::UpdateForWindowSizeChange();
}

void ShapeRenderer::Render()
{
    m_d2dContext->BeginDraw();
    m_d2dContext->Clear(D2D1::ColorF(m_backgroundColor.R/255.0f, m_backgroundColor.G/255.0f, m_backgroundColor.B/255.0f));
    m_d2dContext->SetTransform(D2D1::Matrix3x2F::Identity());

    // Render Beziers
    BezierRender();

    // We ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    const auto hr = m_d2dContext->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    Present();
}

void ShapeRenderer::SaveCanvas(IRandomAccessStream^ randomAccessStream, GUID wicFormat)
{
    // Convert the RandomAccessStream to an IStream.
    ComPtr<IStream> stream;
    DX::ThrowIfFailed(
        CreateStreamOverRandomAccessStream(randomAccessStream, IID_PPV_ARGS(&stream))
        );

    Render();

    // Save the rendered drawing to the IStream.
    SaveBitmapToStream(
        m_d2dTargetBitmap.Get(),
        m_wicFactory.Get(),
        m_d2dContext.Get(),
        wicFormat,
        stream.Get()
        );
}

void ShapeRenderer::OnDataRequested(_In_ DataTransferManager^ sender, _In_ DataRequestedEventArgs^ args)
{
    DataPackage^ package = args->Request->Data;

    // Set the title, description, and text to be shared.
    package->Properties->ApplicationName = "DirectX DrawIt app";
    package->Properties->Title = "A custom drawing for you!";
    package->Properties->Description = "This is a drawing made with the DirectX DrawIt app using C++.";

    // Create an in-memory stream to contain the shared data.
    auto randomAccessStream = ref new InMemoryRandomAccessStream();

    // Create an IStream over that random access stream.
    ComPtr<IStream> stream;
    DX::ThrowIfFailed(
        CreateStreamOverRandomAccessStream(randomAccessStream, IID_PPV_ARGS(&stream))
        );

    Render();

    // Save the bitmap to the stream.
    SaveBitmapToStream(
        m_d2dTargetBitmap.Get(),
        m_wicFactory.Get(),
        m_d2dContext.Get(),
        GUID_ContainerFormatPng,
        stream.Get()
        );

    // Create a stream reference for the random access stream and give it to the data package.
    RandomAccessStreamReference^ streamReference = RandomAccessStreamReference::CreateFromStream(randomAccessStream);
    package->SetBitmap(streamReference);
}

// Save render target bitmap to a stream using WIC.
void ShapeRenderer::SaveBitmapToStream(
    _In_ ID2D1Bitmap1* d2dBitmap,
    _In_ IWICImagingFactory2* wicFactory2,
    _In_ ID2D1DeviceContext* d2dContext,
    _In_ REFGUID wicFormat,
    _In_ IStream* stream
    )
{
    // Create and initialize WIC bitmap encoder.
    ComPtr<IWICBitmapEncoder> wicBitmapEncoder;
    DX::ThrowIfFailed(
        wicFactory2->CreateEncoder(
        wicFormat,
        nullptr,
        &wicBitmapEncoder
        )
        );

    DX::ThrowIfFailed(
        wicBitmapEncoder->Initialize(
        stream,
        WICBitmapEncoderNoCache
        )
        );

    // Create and initialize WIC frame encoder.
    ComPtr<IWICBitmapFrameEncode> wicFrameEncode;
    DX::ThrowIfFailed(
        wicBitmapEncoder->CreateNewFrame(
        &wicFrameEncode,
        nullptr     // No encoder options.
        )
        );

    DX::ThrowIfFailed(
        wicFrameEncode->Initialize(nullptr)
        );

    // Retrieve Direct2D device.
    ComPtr<ID2D1Device> d2dDevice;
    d2dContext->GetDevice(&d2dDevice);

    // Create WIC image encoder.
    ComPtr<IWICImageEncoder> imageEncoder;
    DX::ThrowIfFailed(
        wicFactory2->CreateImageEncoder(
        d2dDevice.Get(),
        &imageEncoder
        )
        );

    DX::ThrowIfFailed(
        imageEncoder->WriteFrame(
        d2dBitmap,
        wicFrameEncode.Get(),
        nullptr     // Use default options.
        )
        );

    DX::ThrowIfFailed(
        wicFrameEncode->Commit()
        );

    DX::ThrowIfFailed(
        wicBitmapEncoder->Commit()
        );

    // Flush all memory buffers to the next-level storage object.
    DX::ThrowIfFailed(
        stream->Commit(STGC_DEFAULT)
        );
}

void ShapeRenderer::OnPointerPressed(
    _In_ Object^ /* sender */,
    _In_ PointerRoutedEventArgs^ args
    )
{
    Windows::UI::Input::PointerPoint^ pointerPoint = args->GetCurrentPoint(nullptr);

    // Make sure no pointer is already inking (we allow only one 'active' pointer at a time) 
    // Make sure pointer is in inking mode (pen and no button, or mouse and left button down)
    if (m_pointerId == -1)
    {        
        m_manipulationPoints->Clear();


        switch (GetPointerEventType(args))
        {
        case PointerEventType::Erase:
            m_manipulationMode = Windows::UI::Input::Inking::InkManipulationMode::Erasing;            
            m_manipulationPoints->Append(pointerPoint->Position);
            break;

        case PointerEventType::Ink:
            m_manipulationMode = Windows::UI::Input::Inking::InkManipulationMode::Inking;
            m_renderingMode = Live; // enter live rendering mode
            // this is the first point of the stroke, begin the stroke with the stroke builder
            m_strokeBuilder->BeginStroke(pointerPoint);
            break;       

        default:
            return; // pointer is neither inking nor erasing nor selecting: do nothing
        }

        Render();

        m_pointerId = pointerPoint->PointerId; // save pointer id so that no other pointer can ink until this one it is released      
    }
}

void ShapeRenderer::OnPointerReleased(
    _In_ Object^ /* sender */,
    _In_ PointerRoutedEventArgs^ args
    )
{
    Windows::UI::Input::Inking::InkStroke^ stroke = nullptr;
    Windows::UI::Input::PointerPoint^ pointerPoint = args->GetCurrentPoint(nullptr);

    // Make sure the event belongs to the pointer that is currently inking
    if (m_pointerId == static_cast<int>(pointerPoint->PointerId))
    {
        m_renderingMode = Bezier;    // exit live rendering mode
        m_pointerId = -1;            // release 'active' pointer so that a new pointer may start inking again

        switch (m_manipulationMode)
        {
        case Windows::UI::Input::Inking::InkManipulationMode::Erasing:
            m_strokeContainer->SelectWithLine(m_manipulationPoints->GetAt(0), pointerPoint->Position);
            m_strokeContainer->DeleteSelected();
            break;

        case Windows::UI::Input::Inking::InkManipulationMode::Inking:
            // this is the last point in the stroke
            stroke = m_strokeBuilder->EndStroke(pointerPoint);

            // add stroke to the stoke container
            m_strokeContainer->AddStroke(stroke);
            break;
        }

        m_manipulationPoints->Clear();

        // re-render the entire scene
        Render();
    }
}

void ShapeRenderer::OnPointerMoved(
    _In_ Object^ /* sender */,
    _In_ PointerRoutedEventArgs^ args
    )
{
    Windows::UI::Input::PointerPoint^ pointerPoint = args->GetCurrentPoint(nullptr);

    // Make sure the event belongs to the pointer that is currently inking
    if (m_pointerId == static_cast<int>(pointerPoint->PointerId))
    {
        if (m_manipulationMode == Windows::UI::Input::Inking::InkManipulationMode::Erasing)
        {
            // Erase ink that intersects line from last point to current point
            Windows::Foundation::Rect invalidateRect = m_strokeContainer->SelectWithLine(m_manipulationPoints->GetAt(0), pointerPoint->Position);
            if (invalidateRect.Height != 0 || invalidateRect.Width != 0)
            {
                m_strokeContainer->DeleteSelected();
                Render();
            }

            // Store current point: it will be the starting point at next pointer update
            m_manipulationPoints->Clear();
            m_manipulationPoints->Append(pointerPoint->Position);
        }
        else
        {
            Windows::Foundation::Point previousPoint(0.0f, 0.0f);
            ID2D1StrokeStyle* strokeStyle = nullptr;
            ID2D1SolidColorBrush* brush = nullptr;
            auto width = 0.0f;

            if (m_manipulationMode == Windows::UI::Input::Inking::InkManipulationMode::Inking)
            {
                // Obtain intermediate points (including the last/current one)
                auto intermediatePoints = args->GetIntermediatePoints(nullptr);

                // Update ink manager with all intermediate points
                int i = intermediatePoints->Size - 1;
                // AppendToStroke returns the last point that was added to the stroke builder.
                // We need to save it because it is the initial point of the new line we want to render.
                previousPoint = m_strokeBuilder->AppendToStroke(intermediatePoints->GetAt(i))->Position;
                for (i = i - 1; i >= 0; i--)
                {
                    m_strokeBuilder->AppendToStroke(intermediatePoints->GetAt(i));
                }

                // Setup drawing attributes for live rendering
                strokeStyle = m_inkStyle.Get();
                brush = m_inkBrush.Get();
                width = m_drawingAttributes->Size.Width;
            }
            // Live rendering
            // First we need to copy the content of the last presented buffer: it contains
            // the Beziers and the lines we rendered at previous pointer moves.
            m_d3dContext->CopyResource(m_currentBuffer.Get(), m_previousBuffer.Get());
            // Then we draw a new line, from the last position of the pointer to its current one.
            m_d2dContext->BeginDraw();
            m_d2dContext->SetTransform(D2D1::Matrix3x2F::Identity());
            m_d2dContext->DrawLine(
                D2D1::Point2F(previousPoint.X, previousPoint.Y), 
                D2D1::Point2F(pointerPoint->Position.X, pointerPoint->Position.Y), 
                brush, 
                width, 
                strokeStyle);

            // We ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
            // is lost. It will be handled during the next call to Present.
            const auto hr = m_d2dContext->EndDraw();
            if (hr != D2DERR_RECREATE_TARGET)
            {
                DX::ThrowIfFailed(hr);
            }

            Present();
        }
    }
}

void ShapeRenderer::OnChangeDrawingAttributes(Windows::UI::Color newStrokeColor, Windows::Foundation::Size sz)
{
    m_drawingAttributes->Color = newStrokeColor;

    if (sz != Windows::Foundation::Size(0.0f, 0.0f))
    {
        m_drawingAttributes->Size = sz;
    }
    // update default drawing attributes
    m_strokeBuilder->SetDefaultDrawingAttributes(m_drawingAttributes.Get());

    // update _inkBrush to reflect drawing color change
    CreateBrush(m_d2dContext.Get(), m_drawingAttributes->Color, &m_inkBrush);
}

void ShapeRenderer::BezierRender()
{
    // render all strokes
    const auto inkStrokes = m_strokeContainer->GetStrokes();
    for (auto i = 0U; i < inkStrokes->Size; ++i)
    {
        // obtain the stroke
        auto stroke = inkStrokes->GetAt(i);

        // convert stroke to geometry
        // geometry is a smart pointer and it is released right after it is used
        // for performance reasons one may consider storing geometry paths instead of auto-releasing them once they are used
        Microsoft::WRL::ComPtr<ID2D1PathGeometry> strokeGeometry;
        ConvertStrokeToGeometry(stroke, &strokeGeometry);

        // stroke's width and color are retrieved from the stroke's drawing attributes
        const float width = stroke->DrawingAttributes->Size.Width;
        // create brush from current stroke
        Microsoft::WRL::ComPtr<ID2D1SolidColorBrush> strokeBrush;
        CreateBrush(m_d2dContext.Get(), stroke->DrawingAttributes->Color, &strokeBrush);

        // render stroke geometry
        m_d2dContext->DrawGeometry(strokeGeometry.Get(), strokeBrush.Get(), width, m_inkStyle.Get());                
    }
}

// convert bezier control points in each segment to a bezier curve in the path geometry
void ShapeRenderer::ConvertStrokeToGeometry(_In_ Windows::UI::Input::Inking::InkStroke^ stroke, _Outptr_ ID2D1PathGeometry** geometry)
{
    // create a geometry path
    DX::ThrowIfFailed(m_d2dFactory->CreatePathGeometry(geometry));

    // create and initialize a geometry sink
    Microsoft::WRL::ComPtr<ID2D1GeometrySink> sink;
    DX::ThrowIfFailed((*geometry)->Open(&sink));
    sink->SetSegmentFlags(D2D1_PATH_SEGMENT_FORCE_ROUND_LINE_JOIN);
    sink->SetFillMode(D2D1_FILL_MODE_ALTERNATE);

    // obtain rendering segments for this stroke
    auto renderingSegments = stroke->GetRenderingSegments();

    // obtain first rendering segment, set as a starting point
    const auto firstSegment = renderingSegments->GetAt(0);
    Windows::Foundation::Point first = firstSegment->Position;
    sink->BeginFigure(D2D1::Point2F(first.X, first.Y), D2D1_FIGURE_BEGIN_FILLED);

    // process all remaining rendering segments, add bezier segment to the geometry path sink
    for (auto i = 1U; i < renderingSegments->Size; ++i)
    {
        // obtain i-th rendering segment for the given stroke
        auto renderingSegment = renderingSegments->GetAt(i);

        // add bezier segment to the geometry path sink
        sink->AddBezier(
            D2D1::BezierSegment(
            D2D1::Point2F(renderingSegment->BezierControlPoint1.X, renderingSegment->BezierControlPoint1.Y),
            D2D1::Point2F(renderingSegment->BezierControlPoint2.X, renderingSegment->BezierControlPoint2.Y),
            D2D1::Point2F(renderingSegment->Position.X, renderingSegment->Position.Y)
            )
            );
    }

    // done with given stroke, end and close geometry sink
    // sink will be automatically released as it is a smart pointer
    sink->EndFigure(D2D1_FIGURE_END_OPEN);
    DX::ThrowIfFailed(sink->Close());
}
