#include "pch.h"
#include "XamlSiS.h"

using namespace Microsoft::WRL;
using namespace Windows::UI::Xaml::Media;
using namespace Windows::UI::Xaml::Media::Imaging;

using namespace DXHelpers;
using namespace SiSSample;

XamlSiS::XamlSiS(SurfaceImageSource^ surfaceImageSource, int width, int height)
{
	CreateDeviceIndependentResources(surfaceImageSource);
	CreateDeviceResources();

	// create the WIC factory
	m_pIWICFactory = nullptr;
	ThrowIfFailed(CoCreateInstance(CLSID_WICImagingFactory,
        nullptr,
        CLSCTX_INPROC_SERVER,
        IID_PPV_ARGS(&m_pIWICFactory)
        ));

	_width = width;
	_height = height;
	
}

XamlSiS::~XamlSiS()
{
}

void XamlSiS::CreateDeviceIndependentResources(SurfaceImageSource^ surfaceImageSource)
{	
	// QI for SurfaceImageSourceNative
	IInspectable* sisInspectable = GetIInspectable(surfaceImageSource);
	ThrowIfFailed(
	sisInspectable->QueryInterface(__uuidof(ISurfaceImageSourceNative), (void **)&m_sisNative)
	);
}
	
void XamlSiS::CreateDeviceResources()
{
	// create D3D device
    UINT flags = D3D11_CREATE_DEVICE_BGRA_SUPPORT; 

#if defined(_DEBUG)    
	flags |= D3D11_CREATE_DEVICE_DEBUG;
#endif

	const std::array<D3D_FEATURE_LEVEL, 6> featureLevels = 
    {
        D3D_FEATURE_LEVEL_11_0,
        D3D_FEATURE_LEVEL_10_1,
        D3D_FEATURE_LEVEL_10_0,
        D3D_FEATURE_LEVEL_9_3,
        D3D_FEATURE_LEVEL_9_2,
        D3D_FEATURE_LEVEL_9_1,
    };
	
	ThrowIfFailed(
	D3D11CreateDevice(
		nullptr,
		D3D_DRIVER_TYPE_HARDWARE,
		nullptr,
		flags,
		featureLevels.data(),
		UINT(featureLevels.size()),
		D3D11_SDK_VERSION,
		&m_d3dDevice,
		nullptr,
		&m_d3dContext
		)
	);	

	// QI for DXGI device
	ComPtr<IDXGIDevice> dxgiDevice;

	ThrowIfFailed(
	m_d3dDevice.As(&dxgiDevice)
	);
	
	// get d2d device context
	ComPtr<ID2D1Device> d2dDevice;
    D2D1CreateDevice(dxgiDevice.Get(), nullptr, &d2dDevice);
	ThrowIfFailed(
	d2dDevice->CreateDeviceContext(D2D1_DEVICE_CONTEXT_OPTIONS_NONE, &m_d2dContext)
	);

	ThrowIfFailed(
	m_sisNative->SetDevice(dxgiDevice.Get())
	);
}

void XamlSiS::DrawD2DImage(String^ fileName)
{	
	// calculate area of update rect
	RECT updateRect; 
	updateRect.left = 0;
    updateRect.top = 0;
    updateRect.right = _width;
    updateRect.bottom = _height;

	POINT offset;
	
	ComPtr<IDXGISurface> surface;

	// begin drawing
	HRESULT beginDrawHR = m_sisNative->BeginDraw(updateRect, &surface, &offset);
	if (beginDrawHR == DXGI_ERROR_DEVICE_REMOVED || beginDrawHR == DXGI_ERROR_DEVICE_REMOVED)
	{
		//re-create the device here
		CreateDeviceResources();		
		return DrawD2DImage(fileName);
	}
	else
	{
		ThrowIfFailed(beginDrawHR);
	}	

	ComPtr<ID2D1Bitmap1> bitmap;
	ThrowIfFailed(
    m_d2dContext->CreateBitmapFromDxgiSurface(surface.Get(), nullptr, &bitmap)
	);

	//decode the file here
	//******************************************************************
	HRESULT hr = S_OK;
	// Step 1: Decode the source image

    // Create a decoder
    ComPtr<IWICBitmapDecoder> pDecoder = nullptr;
	
    hr = m_pIWICFactory->CreateDecoderFromFilename(
        fileName->Data(),                      // Image to be decoded
        nullptr,                            // Do not prefer a particular vendor
        GENERIC_READ,                    // Desired read access to the file
        WICDecodeMetadataCacheOnDemand,  // Cache metadata when needed
        &pDecoder                        // Pointer to the decoder
        );

    // Retrieve the first frame of the image from the decoder
    ComPtr<IWICBitmapFrameDecode> pFrame = nullptr;

    if (SUCCEEDED(hr))
    {
        hr = pDecoder->GetFrame(0, &pFrame);
    }

    //Step 2: Format convert the frame to 32bppPBGRA
    if (SUCCEEDED(hr))
    {		
		m_pConvertedSourceBitmap = nullptr;
        hr = m_pIWICFactory->CreateFormatConverter(&m_pConvertedSourceBitmap);
    }

    if (SUCCEEDED(hr))
    {
        hr = m_pConvertedSourceBitmap->Initialize(
			pFrame.Get(),                          // Input bitmap to convert
            GUID_WICPixelFormat32bppPBGRA,   // Destination pixel format
            WICBitmapDitherTypeNone,         // Specified dither pattern
            nullptr,                            // Specify a particular palette 
            0.f,                             // Alpha threshold
            WICBitmapPaletteTypeCustom       // Palette translation type
            );
    }

	if (SUCCEEDED(hr))
    {
        // Need to release the previous D2DBitmap if there is one
		m_pD2DBitmap = nullptr;		
        hr = m_d2dContext->CreateBitmapFromWicBitmap(m_pConvertedSourceBitmap.Get(), nullptr, &m_pD2DBitmap);
    }
	pDecoder.Reset();
	pFrame.Reset();

	//***************************************************************

	m_d2dContext->BeginDraw();
    m_d2dContext->SetTarget(bitmap.Get());
	
	// clear with background color. Note this will eventually clip to the update
	//  area when the D3D guard rect is implemented
	m_d2dContext->SetTransform(D2D1::Matrix3x2F::Identity());
    m_d2dContext->Clear(D2D1::ColorF(D2D1::ColorF::White));	

	// D2DBitmap may have been released due to device loss. 
    // If so, re-create it from the source bitmap
    if (m_pConvertedSourceBitmap && !m_pD2DBitmap)
    {
        m_d2dContext->CreateBitmapFromWicBitmap(m_pConvertedSourceBitmap.Get(), nullptr, &m_pD2DBitmap);
    }

	// Create a rectangle same size of current window
    D2D1_RECT_F rectangle = D2D1::RectF((FLOAT) offset.x,
        (FLOAT) offset.y,
        (FLOAT) offset.x + (FLOAT) (updateRect.right - updateRect.left),
        (FLOAT) offset.y + (FLOAT) (updateRect.bottom - updateRect.top));

    // Draws an image and scales it to the current window size
    if (m_pD2DBitmap)
    {
		m_d2dContext->DrawBitmap(m_pD2DBitmap.Get(), rectangle);
    }

	// clear target, end drawing
	m_d2dContext->SetTarget(nullptr);

	ThrowIfFailed(
	m_d2dContext->EndDraw()
	);

	ThrowIfFailed(
	m_sisNative->EndDraw()
	);
}
