#include "pch.h"
#include "TriangleRenderer.h"

using namespace DirectX;
using namespace Microsoft::WRL;
using namespace Windows::Foundation;
using namespace Windows::UI::Core;

TriangleRenderer::TriangleRenderer() :
    m_loadingComplete(false),
    m_numVertices(3)
{
}

void TriangleRenderer::CreateDeviceResources()
{
    Direct3DBase::CreateDeviceResources();

    auto loadVSTask = DX::ReadDataAsync("SimpleVertexShader.cso");
    auto loadPSTask = DX::ReadDataAsync("SimplePixelShader.cso");

    auto createVSTask = loadVSTask.then([this](Platform::Array<byte>^ fileData)
    {
        DX::ThrowIfFailed(
            m_d3dDevice->CreateVertexShader(
            fileData->Data,
            fileData->Length,
            nullptr,
            &m_vertexShader
            )
            );

        const std::array<D3D11_INPUT_ELEMENT_DESC, 1> vertexDesc = 
        {
            { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0,  D3D11_INPUT_PER_VERTEX_DATA, 0 },			
        };

        DX::ThrowIfFailed(
            m_d3dDevice->CreateInputLayout(
			vertexDesc.data(),
			UINT(vertexDesc.size()),
            fileData->Data,
            fileData->Length,
            &m_inputLayout
            )
            );
    });

    auto createPSTask = loadPSTask.then([this](Platform::Array<byte>^ fileData)
    {
        DX::ThrowIfFailed(
            m_d3dDevice->CreatePixelShader(
            fileData->Data,
            fileData->Length,
            nullptr,
            &m_pixelShader
            )
            );
    });

    auto createCubeTask = (createPSTask && createVSTask).then([this] ()
    {
        //create the compute shader
        // A triangle 
        const std::array<Vertex2D, 3> vertices =
        {
            DirectX::XMFLOAT2( -0.25f, 0.0f ),
            DirectX::XMFLOAT2( 0.0f, -0.5f ),
            DirectX::XMFLOAT2( -0.5f, -0.5f ),
        };

        // create the AMP compute engine here 
        m_AmpComputeEngine = std::move(std::unique_ptr<AMP_compute_engine>(new AMP_compute_engine(m_d3dDevice.Get())));
		m_AmpComputeEngine->initialize_data(m_numVertices, vertices.data());
        std::array<ID3D11Buffer*, 1> pBuffer = { nullptr };
        HRESULT hr = m_AmpComputeEngine->get_data_d3dbuffer(reinterpret_cast<void**>(pBuffer.data()));
        m_vertexPosBuffer = pBuffer[0];

        // Bind a resource view to the CS buffer
        D3D11_BUFFER_DESC descBuf;
        ZeroMemory( &descBuf, sizeof(descBuf) );
        m_vertexPosBuffer->GetDesc(&descBuf);

        D3D11_SHADER_RESOURCE_VIEW_DESC DescRV;
        ZeroMemory( &DescRV, sizeof( DescRV ) );
        DescRV.Format = DXGI_FORMAT_R32_TYPELESS;
        DescRV.ViewDimension = D3D11_SRV_DIMENSION_BUFFEREX;
        DescRV.BufferEx.Flags = D3D11_BUFFEREX_SRV_FLAG_RAW;
        DescRV.Buffer.FirstElement = 0;
        DescRV.Buffer.NumElements = descBuf.ByteWidth / sizeof(int); 
        ID3D11Resource* pRes = m_vertexPosBuffer.Get();
        m_d3dDevice->CreateShaderResourceView(pRes, &DescRV, &m_VertexPosBufferRV);
        pRes = nullptr;
        //Set the input layout
        m_d3dContext->IASetInputLayout(m_inputLayout.Get());
        D3D11_BUFFER_DESC bd;
        ZeroMemory( &bd, sizeof(bd) );
        bd.Usage = D3D11_USAGE_DEFAULT;
        bd.ByteWidth = sizeof( Vertex2D ) * m_numVertices;
        bd.BindFlags = D3D11_BIND_VERTEX_BUFFER;
        bd.CPUAccessFlags = 0;
        hr = m_d3dDevice->CreateBuffer(&bd, nullptr, &m_vertexBuffer);

        // Set vertex buffer
        UINT stride = sizeof( Vertex2D );
        UINT offset = 0;
        m_d3dContext->IASetVertexBuffers(0, 1, &m_vertexBuffer, &stride, &offset);

        // Set primitive topology
        m_d3dContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST );
    });

    createCubeTask.then([this] ()
    {
        m_loadingComplete = true;
    });
}

void TriangleRenderer::CreateWindowSizeDependentResources()
{
    Direct3DBase::CreateWindowSizeDependentResources();
}

void TriangleRenderer::Update(float timeTotal, float timeDelta)
{
}

//--------------------------------------------------------------------------------------
// Run compute shader to acquire data
//--------------------------------------------------------------------------------------
void TriangleRenderer::ComputeRenderData()
{
    // Only draw the triangle once it is loaded (loading is asynchronous).
    if (!m_loadingComplete)
    {
        return;
    }
    m_AmpComputeEngine->run();
}

void TriangleRenderer::Render()
{
    // Only draw the triangle once it is loaded (loading is asynchronous).
    if (!m_loadingComplete)
    {
        return;
    }

    // Bind the vertex shader data though the compute shader result buffer view
    UINT stride = sizeof( Vertex2D );
    UINT offset = 0;
    m_d3dContext->IASetVertexBuffers(0, 1, m_vertexBuffer.GetAddressOf(), &stride, &offset);
    m_d3dContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST );

    const std::array<ID3D11ShaderResourceView*, 1> aRViews = { m_VertexPosBufferRV.Get() };
	m_d3dContext->VSSetShaderResources( 0, 1, aRViews.data() );

    // Clear the back buffer 
    const std::array<float, 4> ClearColor = { 0.0f, 0.125f, 0.3f, 1.0f };
    m_d3dContext->ClearRenderTargetView(
        m_renderTargetView.Get(),
		ClearColor.data()
        );

    m_d3dContext->OMSetRenderTargets(1, m_renderTargetView.GetAddressOf(), nullptr);

    // Render the triangle
    m_d3dContext->VSSetShader( m_vertexShader.Get(), nullptr, 0 );
    m_d3dContext->PSSetShader( m_pixelShader.Get(), nullptr, 0 );
    m_d3dContext->Draw( m_numVertices, 0 );

    const std::array<ID3D11ShaderResourceView*, 1> ppSRVNULL = { nullptr };
    m_d3dContext->VSSetShaderResources( 0, 1, ppSRVNULL.data() );

    // Present the information rendered to the back buffer to the front buffer (the screen)
    Present();
}
