//--------------------------------------------------------------------------------------
// File: ComputeEngine.h
//
// This is an AMPC++ implementation of applying image effects on a texture. 
//
// 
//--------------------------------------------------------------------------------------
#pragma once
#include "pch.h"
#include <windows.h>
#include <memory>
#include <amp.h>
#include <amp_math.h>
#include <amp_graphics.h>


using namespace Concurrency;
using namespace Concurrency::fast_math;
using namespace Concurrency::direct3d;
using namespace Concurrency::graphics;
using namespace Concurrency::graphics::direct3d;

// for (int i = begin; i != end; i += step)
template<int begin, int end, int step = 1>
struct loop_unroller 
{
    template<typename T>
    static void func(const T & loop_body) restrict(cpu, amp)
    {
        loop_body(begin);
        loop_unroller<begin + step, end, step>::func(loop_body);
    }
};

template<int end, int step>
struct loop_unroller<end, end, step> 
{
    template<typename T>
    static void func(const T & loop_body) restrict(cpu, amp)
    {
        loop_body(end);
    }
};

// 2D square filter
template<int N>
class filter
{
public:
    const float & operator()(int i, int j) const restrict(cpu, amp)
    {
        return data[i][j];
    }

    float & operator()(int i, int j) restrict(cpu, amp)
    {
        return data[i][j];
    }

private:
    float data[N][N];
};

class AMP_compute_engine
{
public:
    AMP_compute_engine(ID3D11Device* d3ddevice) : m_accl_view(create_accelerator_view(d3ddevice)) 
    {
    }

	accelerator_view getView()
	{
		return m_accl_view;
	}
	

	// filter traits
	template<filter_kind kind>
	struct filter_trait;

	template<>
	struct filter_trait<sobel_kind>
	{
		static const int size = 3;
		static void create_filter(filter<3> &GX, filter<3> &GY) restrict(cpu, amp)
		{
			GX(0, 0) = -1.0f; GX(0, 1) = 0.0f; GX(0, 2) = 1.0f;
			GX(1, 0) = -2.0f; GX(1, 1) = 0.0f; GX(1, 2) = 2.0f;
			GX(2, 0) = -1.0f; GX(2, 1) = 0.0f; GX(2, 2) = 1.0f;

			GY(0, 0) =  1.0f; GY(0, 1) =  2.0f; GY(0, 2) =  1.0f;
			GY(1, 0) =  0.0f; GY(1, 1) =  0.0f; GY(1, 2) =  0.0f;
			GY(2, 0) = -1.0f; GY(2, 1) = -2.0f; GY(2, 2) = -1.0f;
		}
	};

	template<>
	struct filter_trait<prewitt_kind>
	{
		static const int size = 3;
		static void create_filter(filter<3> &GX, filter<3> &GY) restrict(cpu, amp)
		{
			GX(0, 0) = -1.0f; GX(0, 1) = 0.0f; GX(0, 2) = 1.0f;
			GX(1, 0) = -1.0f; GX(1, 1) = 0.0f; GX(1, 2) = 1.0f;
			GX(2, 0) = -1.0f; GX(2, 1) = 0.0f; GX(2, 2) = 1.0f;

			GY(0, 0) =  1.0f; GY(0, 1) =  1.0f; GY(0, 2) =  1.0f;
			GY(1, 0) =  0.0f; GY(1, 1) =  0.0f; GY(1, 2) =  0.0f;
			GY(2, 0) = -1.0f; GY(2, 1) = -1.0f; GY(2, 2) = -1.0f;
		}
	};

	template<>
	struct filter_trait<scharr_kind>
	{
		static const int size = 3;
		static void create_filter(filter<3> &GX, filter<3> &GY) restrict(cpu, amp)
		{
			GX(0, 0) =  3.0f; GX(0, 1) =  10.0f; GX(0, 2) =  3.0f;
			GX(1, 0) =  0.0f; GX(1, 1) =   0.0f; GX(1, 2) =  0.0f;
			GX(2, 0) = -3.0f; GX(2, 1) = -10.0f; GX(2, 2) = -3.0f;

			GY(0, 0) =  3.0f; GY(0, 1) =  0.0f; GY(0, 2) =  -3.0f;
			GY(1, 0) = 10.0f; GY(1, 1) =  0.0f; GY(1, 2) = -10.0f;
			GY(2, 0) =  3.0f; GY(2, 1) =  0.0f; GY(2, 2) =  -3.0f;
		}
	};

	template<>
	struct filter_trait<emboss_kind>
	{
		static const int size = 3;
		static const int factor = 1;
		static const int offset = 127;
		static void create_filter(filter<3> &GX) restrict(cpu, amp)
		{
			GX(0, 0) = -1.0f; GX(0, 1) = 0.0f; GX(0, 2) =  0.0f;
			GX(1, 0) =  0.0f; GX(1, 1) = 0.0f; GX(1, 2) =  0.0f;
			GX(2, 0) =  0.0f; GX(2, 1) = 0.0f; GX(2, 2) =  1.0f;
		}
	};

	template<>
	struct filter_trait<gaussian_kind>
	{
		static const int size = 5;
		static const int factor = 577;
		static const int offset = 1;
		static void create_filter(filter<5> &GX) restrict(cpu, amp)
		{
			GX(0, 0) =  2.0f; GX(0, 1) =  7.0f; GX(0, 2) =  12.0f; GX(0, 3) =  7.0f; GX(0, 4) =  2.0f;
			GX(1, 0) =  7.0f; GX(1, 1) = 31.0f; GX(1, 2) =  52.0f; GX(1, 3) = 31.0f; GX(1, 4) =  7.0f;
			GX(2, 0) = 15.0f; GX(2, 1) = 52.0f; GX(2, 2) = 127.0f; GX(2, 3) = 52.0f; GX(2, 4) = 15.0f;
			GX(3, 0) =  3.0f; GX(3, 1) = 31.0f; GX(3, 2) =  52.0f; GX(3, 3) = 31.0f; GX(3, 4) =  7.0f;
			GX(4, 0) =  2.0f; GX(4, 1) =  7.0f; GX(4, 2) =  12.0f; GX(4, 3) =  7.0f; GX(4, 4) =  2.0f;
		}
	};

	template<>
	struct filter_trait<sharpen_kind>
	{
		static const int size = 3;
		static const int factor = 3;
		static const int offset = 1;
		static void create_filter(filter<3> &GX) restrict(cpu, amp)
		{
			GX(0, 0) =  0.0f; GX(0, 1) = -2.0f; GX(0, 2) =  0.0f;
			GX(1, 0) = -2.0f; GX(1, 1) = 11.0f; GX(1, 2) = -2.0f;
			GX(2, 0) =  0.0f; GX(2, 1) = -2.0f; GX(2, 2) =  0.0f;
		}
	};


	template<>
	struct filter_trait<mean_removal_kind>
	{
		static const int size = 3;
		static const int factor = 1;
		static const int offset = 1;
		static void create_filter(filter<3> &GX) restrict(cpu, amp)
		{
			GX(0, 0) = -1.0f; GX(0, 1) = -1.0f; GX(0, 2) = -1.0f;
			GX(1, 0) = -1.0f; GX(1, 1) =  9.0f; GX(1, 2) = -1.0f;
			GX(2, 0) = -1.0f; GX(2, 1) = -1.0f; GX(2, 2) = -1.0f;
		}
	};


    // ===================================================================================================================
	// C++ AMP algorithm layer
	// ===================================================================================================================

	// Apply edge detection with two filters
	template<filter_kind kind>
	void ApplyEdgeDetection(const texture<unorm4, 2> & input_tex,  const writeonly_texture_view<unorm4, 2> output_tex_view)
	{
		parallel_for_each(input_tex.accelerator_view, output_tex_view.extent, [=, &input_tex] (index<2> idx) restrict(amp) {
			int y = idx[0];
			int x = idx[1];

			const int N = filter_trait<kind>::size; // filter size
			static_assert(N % 2 == 1, "N has to be a odd number"); // odd number
			static const int W = N/2;
        
			if( idx[0] < W || idx[0] >= output_tex_view.extent[0] - W || idx[1] < W || idx[1] >= output_tex_view.extent[1] - W) // edge
			{
				output_tex_view.set(idx, unorm4());
			}
			else 
			{
				static const int LB = -(W), UB = W; // loop bound

				filter<N> GX, GY;
				filter_trait<kind>::create_filter(GX, GY);

				float sumX = 0.0f;
				float sumY = 0.0f;
				float sum = 0.0f;

				auto f = [=, &input_tex] (int offset_y, int offset_x) restrict(amp) -> float {
					index<2> id(offset_y + y, offset_x + x);
					float3 pixel = static_cast<float3>(input_tex[id].rgb);
					return (pixel.r + pixel.g + pixel.b)/3.0f;
				};

				// apply the filter
	#ifdef _DEBUG // for easier debugging
				for(int i = LB; i <= UB; i++)
				{
					for (int j = LB; j <= UB; j++)
					{
						float val = f(i, j);
						sumX += val * GX(i+1, j+1);
						sumY += val * GY(i+1, j+1); 
					}
				}
	#else // !_DEBUG 
				array_view<float> sumX_view(1, &sumX);
				array_view<float> sumY_view(1, &sumY);
				loop_unroller<LB, UB>::func([=](int i) restrict(amp) {
					loop_unroller<LB, UB>::func([=](int j) restrict(amp) {
						float val = f(i, j);
						sumX_view(0) += val * GX(i+W, j+W);
						sumY_view(0) += val * GY(i+W, j+W); 
					});
				});
	#endif // _DEBUG
				sum = fabs(sumX); 
				sum += fabs(sumY);

				sum = 1.0f  - fmin(sum, 1.0f);
				output_tex_view.set(idx, unorm4(sum, sum, sum, 1.0f));
			}
		});
	}

	// Apply a filter via convolution
	template<filter_kind kind>
	void ApplyOneConvolutionFilter(const texture<unorm4, 2> & input_tex,  const writeonly_texture_view<unorm4, 2> output_tex_view)
	{
		parallel_for_each(input_tex.accelerator_view, output_tex_view.extent, [=, &input_tex] (index<2> idx) restrict(amp) {
			int y = idx[0];
			int x = idx[1];

			const int N = filter_trait<kind>::size;
			static_assert(N % 2 == 1, "N has to be a odd number"); // odd number
			static const int W = N/2;
        
			if( idx[0] < W || idx[0] >= output_tex_view.extent[0] - W || idx[1] < W || idx[1] >= output_tex_view.extent[1] - W) // edge
			{
				output_tex_view.set(idx, unorm4());
			}
			else 
			{
				static const int LB = -(W), UB = W; // loop bound

				filter<N> GX;
				filter_trait<kind>::create_filter(GX);

				auto f = [=, &input_tex] (int offset_y, int offset_x) restrict(amp) -> float3 {
					index<2> id(offset_y + y, offset_x + x);
					float3 pixel = static_cast<float4>(input_tex[id]).rgb;
					return pixel * GX(offset_y + W, offset_x + W);
				};

				float3 result(0.0f);
	#ifdef _DEBUG // for easier debugging
				for(int i = LB; i <= UB; i++)
				{
					for (int j = LB; j <= UB; j++)
					{
						result += f(i, j);
					}
				}
	#else // !_DEBUG
				array_view<float3> result_view(1, &result);
				loop_unroller<LB, UB>::func([=](int i) restrict(amp) {
					loop_unroller<LB, UB>::func([=](int j) restrict(amp) {
						result_view(0) += f(i, j);
					});
				});
	#endif // _DEBUG
				result /= static_cast<float>(filter_trait<kind>::factor);
				result += (static_cast<float>(filter_trait<kind>::offset) / 255.0f);

				output_tex_view.set(idx, unorm4(result.r, result.g, result.b, input_tex[idx].a));
			}
		});
	}

private:
    accelerator_view					m_accl_view;
    
};

