// MatrixMul.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

void ComputeMatrixMult(array<float, 2> &mC, array<float, 2> &mA, array<float, 2> &mB);

//helper functions to warm up the runtime and JIT the kernel
void WarmUp()
{
    extent<2> eA(1, 1), eB(1, 1), eC(1, 1);
    array<float, 2> mA(eA), mB(eB), mC(eC);
    ComputeMatrixMult(mC, mA, mB);
}

void DoMatrixMulWithoutAmp(std::vector<int>& vC, const std::vector<int>& vA, const std::vector<int>& vB, int M, int N, int W)
{
	// computes result serially	
	for(int i = 0; i< M; ++i)
	{
		for(int j = 0; j < N; ++j)
		{
			int sum = 0;
			for (int k = 0; k < W; ++k)
				sum += vA[i * W + k] * vB[k * N + j];
			vC[i * N + j] = sum;
		}
	}		
}

void DoMatrixMulWithAmp(std::vector<int>& vC, const std::vector<int>& vA, const std::vector<int>& vB, int M, int N, int W)
{	
	array_view<const int, 2> a(M, W, vA), b(W, N, vB);
    array_view<int, 2> c(M, N, vC);
	c.discard_data();

    parallel_for_each(c.extent, [=](index<2> idx) restrict(amp) 
    {
        int row = idx[0]; int col = idx[1];
        int sum = 0;
        for(int i = 0; i < b.extent[0]; i++) 
            sum += a(row, i) * b(i, col);
        c[idx] = sum;
    });  
	c.synchronize();
}

void DoMatrixMulWithAmpTiling(std::vector<int>& vC, const std::vector<int>& vA, const std::vector<int>& vB, int M, int N, int W)
{		
	static const int TS = 16;
	array_view<const int, 2> a(M, W, vA), b(W, N, vB);
    array_view<int, 2> c(M, N, vC);
	c.discard_data();

	parallel_for_each(c.extent.tile<TS, TS>(), [=](tiled_index<TS, TS> t_idx) restrict(amp) 
    {
		index<2> idx = t_idx.global;
        int row = idx[0]; int col = idx[1];
        int sum = 0;
        for(int i = 0; i < b.extent[0]; i++) 
            sum += a(row, i) * b(i, col);
        c[idx] = sum;
    });
	c.synchronize();
}

void DoMatrixMulwithAmpTileBarrier(std::vector<int>& vC, const std::vector<int>& vA, const std::vector<int>& vB, int M, int N, int W)
{	

	static const int TS = 16;
	array_view<const int, 2> a(M, W, vA), b(W, N, vB);
    array_view<int, 2> c(M, N, vC);
	c.discard_data();

	parallel_for_each(c.extent.tile< TS, TS >(),
	   [=] (tiled_index< TS, TS> t_idx) restrict(amp) 
	   {
		 int row = t_idx.local[0]; int col = t_idx.local[1];
		 int sum = 0;

		 for (int i = 0; i < W; i += TS) {
			tile_static int locA[TS][TS], locB[TS][TS];
			locA[row][col] = a(t_idx.global[0], col + i);
			locB[row][col] = b(row + i, t_idx.global[1]);
			//need to synchronize threads in each tile and hence we need a call to wait on the barrier
			t_idx.barrier.wait();
			for (int k = 0; k < TS; k++)
			  sum += locA[row][k] * locB[k][col];
			// need a wait call on the barrier to synchronize the threads
			t_idx.barrier.wait();
		}
		 c[t_idx.global] = sum;
	   });	
	c.synchronize();
}

void GenerateMatrices()
{
	// rows and columns for matrix
    const int M = 1024;
    const int N = 1024;
    const int W = 1024;

    // create storage for a matrix of above size    
    std::vector<int> vA(M * W); 
    std::vector<int> vB(W * N);

    // Populate matrix objects
    int i = 0;
    std::generate(vA.begin(), vA.end(), [&i](){return i++;});
    std::generate(vB.begin(), vB.end(), [&i](){return i--;});

    // output storage for matrix calculation
    std::vector<int> vC(M * N);

	auto t1 = std::chrono::high_resolution_clock::now();
	DoMatrixMulWithoutAmp(vC, vA, vB, M, N, W);
	auto t2 = std::chrono::high_resolution_clock::now();
	
	std::wcout<<"The matrix multiplication without C++ AMP took "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count()<< " milliseconds to complete"<<std::endl;

	t1 = std::chrono::high_resolution_clock::now();
	DoMatrixMulWithAmp(vC, vA, vB, M, N, W);
	t2 = std::chrono::high_resolution_clock::now();
	
	std::wcout<<"The matrix multiplication with C++ AMP Simple Model took "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count()<< " milliseconds to complete"<<std::endl;

	t1 = std::chrono::high_resolution_clock::now();
	DoMatrixMulWithAmpTiling(vC, vA, vB, M, N, W);
	t2 = std::chrono::high_resolution_clock::now();
	
	std::wcout<<"The matrix multiplication with C++ AMP tiling took "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count()<< " milliseconds to complete"<<std::endl;

	t1 = std::chrono::high_resolution_clock::now();
	DoMatrixMulwithAmpTileBarrier(vC, vA, vB, M, N, W);
	t2 = std::chrono::high_resolution_clock::now();
	
	std::wcout<<"The matrix multiplication with C++ AMP tiling barrier took "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count()<< " milliseconds to complete"<<std::endl;
}

void ComputeMatrixMult(array<float, 2> &mC,
                        array<float, 2> &mA,
                        array<float, 2> &mB)
{
    parallel_for_each(mC.extent,
                      [&](index<2> idx) restrict(amp) {
    
        float result = 0.0f;
        for(int i=0; i<mA.extent[1]; ++i)
        {
            index<2> idxA(idx[0], i);
            index<2> idxB(i, idx[1]);

            result += mA[idxA] * mB[idxB];
        }

        mC[idx] = result;
    });
}

int main()
{
	WarmUp();
	GenerateMatrices();
	std::wcout<< "Hit any key to exit"<<std::endl;
	return 0;
}

