﻿//
// MainPage.xaml.cpp
// Implementation of the MainPage class.
//

#include "pch.h"
#include "MainPage.xaml.h"
#include <chrono>
#pragma warning ( disable : 4267)
using namespace XamlBitonicSort;

using namespace Platform;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::UI::Xaml;
using namespace Windows::UI::Xaml::Controls;
using namespace Windows::UI::Xaml::Controls::Primitives;
using namespace Windows::UI::Xaml::Data;
using namespace Windows::UI::Xaml::Input;
using namespace Windows::UI::Xaml::Media;
using namespace Windows::UI::Xaml::Navigation;

// The Blank Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=234238
unsigned int length = NUM_ELEMENTS;

MainPage::MainPage()
{
	InitializeComponent();
}

/// <summary>
/// Invoked when this page is about to be displayed in a Frame.
/// </summary>
/// <param name="e">Event data that describes how this page was reached.  The Parameter
/// property is typically used to configure the page.</param>
void MainPage::OnNavigatedTo(NavigationEventArgs^ e)
{
	(void) e;	// Unused parameter
}

//----------------------------------------------------------------------------
// Kernel implements partial sorting on accelerator, BITONIC_BLOCK_SIZE at a time
//----------------------------------------------------------------------------
template <typename _type>
void bitonic_sort_kernel(array<_type, 1>& data, unsigned ulevel, unsigned ulevelmask, tiled_index<BITONIC_BLOCK_SIZE> tidx) restrict (amp)
{
    tile_static _type sh_data[BITONIC_BLOCK_SIZE];

    int local_idx = tidx.local[0];
    int global_idx = tidx.global[0];

    // Cooperatively load data - each thread will load data from global memory
    // into tile_static
    sh_data[local_idx] = data[global_idx];

    // Wait till all threads have loaded their portion of data
    tidx.barrier.wait();
    
    // Sort data in tile_static memory
    for (unsigned j = ulevel >> 1 ; j > 0 ; j >>= 1)
    {
        _type result = ((sh_data[local_idx & ~j] <= sh_data[local_idx | j]) == (bool)(ulevelmask & global_idx)) ? sh_data[local_idx ^ j] : sh_data[local_idx];
        tidx.barrier.wait();
        sh_data[local_idx] = result;
        tidx.barrier.wait();
    }
    
    // Store shared data
    data[global_idx] = sh_data[local_idx];
}

//----------------------------------------------------------------------------
// Kernel implements 2D matrix transpose
//----------------------------------------------------------------------------
template <typename _type>
void transpose_kernel(array<_type, 1>& data_in, array<_type, 1>& data_out, unsigned width, unsigned height, tiled_index<TRANSPOSE_BLOCK_SIZE, TRANSPOSE_BLOCK_SIZE> tidx) restrict (amp)
{
	tile_static _type transpose_shared_data[TRANSPOSE_BLOCK_SIZE][TRANSPOSE_BLOCK_SIZE];
	extent<2> e_mat_dim(width, height);

    array_view<_type, 2> transpose_matrix_a = data_in.view_as(e_mat_dim);
	transpose_shared_data[tidx.local[0]][tidx.local[1]] = transpose_matrix_a[tidx.global[0]][tidx.global[1]];

    tidx.barrier.wait();

	array_view<_type, 2> transpose_matrix_b = data_out.view_as(e_mat_dim);
	transpose_matrix_b[tidx.global[1]][tidx.global[0]] = transpose_shared_data[tidx.local[0]][tidx.local[1]];
}


//----------------------------------------------------------------------------
// CPU helper driving accelerator sorting
//----------------------------------------------------------------------------
template <typename _type>
void bitonic_sort_amp(std::vector<_type>& data_in, std::vector<_type>& data_out)
{
	// Verify assumptions
	assert(NUM_ELEMENTS/MATRIX_WIDTH == MATRIX_WIDTH);
	assert(((MATRIX_WIDTH%TRANSPOSE_BLOCK_SIZE) == 0) && ((MATRIX_HEIGHT%TRANSPOSE_BLOCK_SIZE) == 0));
	assert(data_out.size() == NUM_ELEMENTS);

    array<_type, 1> temp(data_out.size());
    array<_type, 1> data(data_out.size(), data_in.begin());

    // Sort the data
    // First sort the rows for the levels <= to the block size
    extent<1> compute_domain(NUM_ELEMENTS);

    for (unsigned level = 2; level <= BITONIC_BLOCK_SIZE ; level = level * 2 )
    {
        parallel_for_each(compute_domain.tile<BITONIC_BLOCK_SIZE>(), [=, &data] (tiled_index<BITONIC_BLOCK_SIZE> tidx) restrict(amp)
        {
            bitonic_sort_kernel<_type>(data, level, level, tidx);
        });
    }

    unsigned ulevel;
    unsigned ulevelMask;
    unsigned width;
    unsigned height;

    // Then sort the rows and columns for the levels > than the block size
    // Transpose. Sort the Columns. Transpose. Sort the Rows.
    for( unsigned level = (BITONIC_BLOCK_SIZE * 2) ; level <= NUM_ELEMENTS ; level = level * 2 )
    {
        ulevel = (level / BITONIC_BLOCK_SIZE);
        ulevelMask = (level & ~NUM_ELEMENTS) / BITONIC_BLOCK_SIZE;
        width = MATRIX_WIDTH;
        height = MATRIX_HEIGHT;

        // Transpose the data from buffer 1 into buffer 2
        extent<2> cdomain_transpose(MATRIX_WIDTH, MATRIX_HEIGHT);
        parallel_for_each (cdomain_transpose.tile<TRANSPOSE_BLOCK_SIZE, TRANSPOSE_BLOCK_SIZE>(), 
            [=, &data, &temp] (tiled_index<TRANSPOSE_BLOCK_SIZE, TRANSPOSE_BLOCK_SIZE> tidx)  restrict(amp)
            {
                transpose_kernel<_type>(data, temp, width, height, tidx);
            });

        // Sort the transposed column data
        extent<1> cdomain_num_elements(NUM_ELEMENTS);
        parallel_for_each(cdomain_num_elements.tile<BITONIC_BLOCK_SIZE>(), [=, &temp] (tiled_index<BITONIC_BLOCK_SIZE> tidx) restrict(amp)
        {
            bitonic_sort_kernel<_type>(temp, ulevel, ulevelMask, tidx);
        });

        ulevel = BITONIC_BLOCK_SIZE;
        ulevelMask = level;
        width = MATRIX_HEIGHT;
        height = MATRIX_WIDTH;

        // Transpose the data from buffer 2 back into buffer 1
        parallel_for_each (cdomain_transpose.tile<TRANSPOSE_BLOCK_SIZE, TRANSPOSE_BLOCK_SIZE>(), 
            [=, &data, &temp] (tiled_index<TRANSPOSE_BLOCK_SIZE, TRANSPOSE_BLOCK_SIZE> tidx)   restrict(amp)
            {
                transpose_kernel<_type>(temp, data, width, height, tidx);
            });

        // Sort the row data
        parallel_for_each(cdomain_num_elements.tile<BITONIC_BLOCK_SIZE>(), [=, &data] (tiled_index<BITONIC_BLOCK_SIZE> tidx) restrict(amp)
        {
            bitonic_sort_kernel<_type>(data, ulevel, ulevelMask, tidx);
        });
    }

	copy(data, data_out.begin());
}


//----------------------------------------------------------------------------
// Generate random data
//----------------------------------------------------------------------------
template <typename _type>
void filldata(std::vector<_type>& data)
{
    for (unsigned i = 0; i < data.size(); i++)
        data[i] = (_type)(rand());
}

template <typename _type>
bool verify(std::vector<_type>& data)
{
    _type max_so_far = data[0];
    for (unsigned i = 1; i < data.size(); i++)
    {
        if (data[i] < max_so_far)
        {
            printf ("Fail. Index %d\n", i);
            return false;
        }
        max_so_far = data[i];
    }
    return true;
}

void XamlBitonicSort::MainPage::SortBtn_Click_1(Platform::Object^ sender, Windows::UI::Xaml::RoutedEventArgs^ e)
{		
	accelerator default_device;
	String^ statusText = ref new String(default_device.get_description().data());
	StatusText->Text = "Using device " ;
	StatusText->Text += statusText;

	if (default_device == accelerator(accelerator::direct3d_ref))
		StatusText->Text = "WARNING!! Running on very slow emulator! Only use this accelerator for debugging.";
	std::vector<int> datain(length);
	std::vector<int> dataOut(length);
	StatusText->Text = "Filling with integer type of length ";
	StatusText->Text += length.ToString();
		
	filldata<int>(datain);
	StatusText->Text = "Offloading to accelerator";

	auto t1 = std::chrono::high_resolution_clock::now();
	bitonic_sort_amp<int>(datain, dataOut);
	StatusText->Text = "Verify data on the CPU";
	if (verify<int>(dataOut))
	{
		auto t2 = std::chrono::high_resolution_clock::now();		
		StatusText->Text = "Verified data on CPU and the sort succeeded. ";
		StatusText->Text += "The Sorting operation took ";
		StatusText->Text += std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count().ToString() + " milliseconds";
		StatusText->Text += "\n";
		StatusText->Text += "The sort data size is " + length.ToString();
	}
	else
	{
		auto t2 = std::chrono::high_resolution_clock::now();		
		StatusText->Text = "Verified data on CPU and the sort failed. ";
		StatusText->Text += "The Sorting operation took ";
		StatusText->Text += std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count().ToString() + " milliseconds";
		StatusText->Text += "\n";
		StatusText->Text += "The sort data size is " + length.ToString();
	}
}
